package boids;

import io.jbotsim.core.Node;
import io.jbotsim.core.Point;
import io.jbotsim.ui.icons.Icons;

import java.util.ArrayList;
import java.util.List;
/**
 * A boid in the flock.
 */
public class Boid extends Node{

    protected Vector2D velocity, nextVelocity;
    protected ArrayList<Boid> theNeighbors;

    /**
     * Initialises the bird with a random direction
     */
    @Override
    public void onStart(){
        velocity = new Vector2D();
        nextVelocity = new Vector2D();
        this.theNeighbors = new ArrayList<Boid>();
        setIcon("/boids/bird.png");
        setIconSize(30);
        setSensingRange(150);
        velocity.setDirection(
                ((Math.random()*4.)-2.),
                ((Math.random()*4.)-2.));
    }

    @Override // Look around
    public void onPreClock(){
        theNeighbors.clear();
        List<Node> sensedNode = getSensedNodes();
        for (Node aNode : sensedNode){
            if (aNode instanceof Boid)
                this.theNeighbors.add((Boid) aNode);
        }
    }

    @Override // Compute movement
    public void onClock(){
        nextVelocity = velocity;
        nextVelocity.add(groupFlock());
        nextVelocity.add(collisionAvoidance());
        nextVelocity.add(matchVelocity());
        nextVelocity.add(new Vector2D(0.02,0.0)); // Wind
        nextVelocityMax(); // Animal limitation
    }

    @Override // Move
    public void onPostClock(){
        Point target = new Point();
        velocity = nextVelocity;
        target.setLocation(getLocation().getX()+velocity.getX(),
                getLocation().getY()+velocity.getY());
        setDirection(target);
        move(velocity.distance());
        wrapLocation();
    }

    /**
     * Rule 1: Boids try to fly towards the centre of mass of
     * neighbouring boids.
     *
     * @return Vector2D
     */
    public Vector2D groupFlock(){
        Vector2D center = new Vector2D();
        for (Boid aBoid : theNeighbors) {
            if(!aBoid.equals(this)) {
                center.add(aBoid.getLocation().getX(),aBoid.getLocation().getY());
            }
        }
        if (theNeighbors.size()>0){
            center.division(theNeighbors.size());
            center.subtract(getLocation().getX(),getLocation().getY());
            center.division(1000.);
        }
        return center;
    }

    /**
     * Rule 2: Boids try to keep a small distance away from other
     * objects (including other boids).
     *
     * @return Vector2D
     */
    public Vector2D collisionAvoidance(){
        Vector2D avoid = new Vector2D();
        for (Boid aBoid : theNeighbors) {
            if(!aBoid.equals(this)) {
                if(distance(aBoid)<60){
                    Vector2D tempDistance = new Vector2D(
                            aBoid.getLocation().getX()-getLocation().getX(),
                            aBoid.getLocation().getY()-getLocation().getY());
                    avoid.subtract(tempDistance);
                }
            }
        }
        avoid.division(20.);
        return avoid;
    }

    /**
     * Rule 3: Boids try to match velocity with near boids.
     *
     * @return Vector2D
     */
    public Vector2D matchVelocity(){
        Vector2D global = new Vector2D();
        for (Boid aBoid : theNeighbors) {
            if(!aBoid.equals(this)) {
                global.add(aBoid.getVelocity());
            }
        }
        if (theNeighbors.size()>0){
            global.division(theNeighbors.size());
            global.subtract(velocity);
            global.division(8.);
        }
        return global;
    }

    private void nextVelocityMax(){
        double tempVelocity = nextVelocity.distance();
        double max = 10;
        if(tempVelocity>max)
            nextVelocity.division(Math.abs(tempVelocity)/max);
    }

    public Vector2D getVelocity(){
        return velocity;
    }

    public boolean visionSight(Node aNode){
        boolean inVisionSight=false;
        Node temp = new Node();
        temp.setLocation(this.getLocation());
        temp.setDirection(aNode.getLocation());
        double delta=velocity.angle() - temp.getDirection();
        if(Math.abs(delta%Math.PI)<2.5){
            inVisionSight=true;
        }
        return inVisionSight;
    }
}
